from typing import Optional

import trio
from pydantic import Field

from ..compute import Value
from ..utils import allow_singular_none, async_cached, url_version
from .installer import make_installer
from .manual_version import ManualVersion

INSTALLER_LOCK = trio.Lock()


class SystemPackage(ManualVersion):
    name_: str = Field(..., alias="name")
    url_: Optional[Value] = Field(default=None, alias="url")
    auto_updates: bool = False
    services: list[str] = Field(default_factory=list, alias="service")
    services_val = allow_singular_none("services")

    async def installer(self):
        return await make_installer(self.driver)

    @property
    def name(self) -> str:
        return self.name_

    @async_cached
    async def url(self) -> Optional[str]:
        if self.url_:
            return await self.url_.compute()
        return None

    async def get_remote_version(self) -> str:
        if url := await self.url():
            return url_version(url)
        if self.auto_updates:
            return "latest"
        return await (await self.installer()).latest_version(self.name)

    async def local_version(self) -> Optional[str]:
        if await self.url():
            return self.cached_version
        installer = await self.installer()
        version = await installer.installed_version(self.name)
        if self.auto_updates:
            return "latest" if version else None
        return version

    async def postinstall_linux(self):
        if self.services:
            await self.driver.with_root(True).run("systemctl", "daemon-reload")
            for service in self.services:
                await self.driver.with_root(True).run("systemctl", "enable", service)

    async def postinstall_macos(self):
        pass

    async def install(self) -> None:
        async with INSTALLER_LOCK:
            installer = await self.installer()
            if url := await self.url():
                await installer.install(url)
                await self.cache_version()
            elif await self.local_version():
                await installer.upgrade(self.name)
            else:
                await installer.install(self.name)
        await (await self.driver.os()).switch(
            linux=self.postinstall_linux, macos=self.postinstall_macos
        )()

        await super().install()



import os
from dataclasses import dataclass
from subprocess import CalledProcessError
from typing import Any, Callable, Iterator, Optional

import requests

from ..driver import OS
from ..filters import Filters, choose
from ..utils import async_cached, async_cached_lock, run_output
from .archive import ArchivePackage


@async_cached_lock
async def github_auth_token() -> Optional[str]:
    try:
        return os.environ["GITHUB_TOKEN"]
    except KeyError:
        pass

    try:
        return await run_output("gh", "auth", "token", silent=True)
    except (CalledProcessError, FileNotFoundError):
        pass

    return None


async def github_api(url: str) -> Any:
    token = await github_auth_token()

    headers = {}
    if token:
        headers["Authorization"] = f"token {token}"

    result = requests.get(f"https://api.github.com/{url}", headers=headers)
    result.raise_for_status()

    return result.json()


@dataclass
class GitHubReleaseArtifact:
    name: str
    url: str


@dataclass
class GitHubRelease:
    id: int
    tag_name: str
    assets: list[GitHubReleaseArtifact]


class GitHubPackage(ArchivePackage, Filters):
    repo: str

    @async_cached
    async def latest_release(self) -> GitHubRelease:
        latest = await github_api(f"repos/{self.repo}/releases/latest")
        return GitHubRelease(
            id=latest["id"],
            tag_name=latest["tag_name"],
            assets=[
                GitHubReleaseArtifact(
                    name=result["name"], url=result["browser_download_url"]
                )
                for result in latest["assets"]
            ],
        )

    @classmethod
    def environment_filters(cls, target_os: OS) -> Iterator[Callable[[str], bool]]:
        for hint in [".tar.gz"]:
            yield cls.includes_(hint)
        for signature_hint in [".asc", ".sig", "sha256", "sha512", ".yml"]:
            yield cls.excludes_(signature_hint)
        for other_os_hint in [".exe", ".dmg"]:
            yield cls.excludes_(other_os_hint)
        for os_hint in target_os.switch(
            linux=[
                cls.includes_("linux"),
                cls.includes_("gnu"),
                cls.excludes_("musl"),
            ],
            macos=[cls.includes_(hint) for hint in ["macos", "darwin", "osx"]],
        ):
            yield os_hint
        arch_hints = ["x86_64", "amd64"]
        for hint in arch_hints:
            yield cls.includes_(hint)

    def all_filters(self, target_os: OS) -> Iterator[Callable[[str], bool]]:
        yield from self.filters()
        yield from self.environment_filters(target_os)

    async def artifact(self) -> GitHubReleaseArtifact:
        candidates = (await self.latest_release()).assets

        def candidate_filter(
            name_filter: Callable[[str], bool]
        ) -> Callable[[GitHubReleaseArtifact], bool]:
            return lambda candidate: name_filter(candidate.name)

        target_os = await self.driver.os()
        return choose(candidates, map(candidate_filter, self.all_filters(target_os)))

    async def archive_url(self) -> str:
        return (await self.artifact()).url

    @property
    def name(self) -> str:
        return self.repo

    async def get_remote_version(self) -> str:
        release = await self.latest_release()
        return str(release.id)



import json
from abc import ABC, abstractmethod
from typing import Any

import requests
from bs4 import BeautifulSoup
from jsonpath_ng import JSONPath as JSONPathT  # type: ignore
from jsonpath_ng import parse as jsonpath_parse  # type: ignore
from pydantic import BaseModel, validator

from .filters import Filters, choose


class Value(BaseModel, ABC):
    @classmethod
    def __get_validators__(cls):
        yield cls.parse

    @staticmethod
    def parse(value: Any) -> "Value":
        if isinstance(value, str):
            return Const(value=value)
        if isinstance(value, dict):
            if base := value.pop("url", None):
                return URL(base=base)
            if base := value.pop("links", None):
                return HTMLLinks(base=base, **value)
            if "format" in value:
                return Format(**value)
            if "jsonpath" in value:
                return JSONPath(**value)
        raise ValueError(f"Cannot parse URL from {value!r}.")

    @abstractmethod
    async def compute(self) -> str:
        raise NotImplementedError


class Const(Value):
    value: str

    async def compute(self):
        return self.value


class Derived(Value, ABC):
    base: Value

    @abstractmethod
    async def derived_value(self, contents: str) -> str:
        raise NotImplementedError

    async def compute(self) -> str:
        base = await self.base.compute()
        return await self.derived_value(base)


class URL(Derived):
    async def derived_value(self, contents: str) -> str:
        return requests.get(contents).text


class JSONPath(Derived, Filters):
    class Config:
        arbitrary_types_allowed = True

    jsonpath: JSONPathT

    @validator("jsonpath", pre=True)
    def parse_jsonpath(cls, value):  # pylint:disable=no-self-argument
        return jsonpath_parse(value)

    async def derived_value(self, contents: str) -> str:
        json_contents = json.loads(contents)
        candidates = [
            candidate.value for candidate in self.jsonpath.find(json_contents)
        ]
        return choose(candidates, self.filters())


class Format(Derived):
    format: str  # pylint:disable=redefined-builtin

    async def derived_value(self, contents: str) -> str:
        return self.format.format(contents)


class HTMLLinks(Derived, Filters):
    async def derived_value(self, contents: str) -> str:
        soup = BeautifulSoup(contents, "html.parser")
        candidates = [link.get("href") for link in soup.find_all("a")]
        return choose(candidates, self.filters())



import subprocess
from functools import wraps
from pathlib import Path
from typing import (
    Any,
    Callable,
    Coroutine,
    Iterable,
    Optional,
    TypeVar,
    Union,
    overload,
)

import requests
import trio
from pydantic import validator

T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")


Some = Optional[Union[T, list[T]]]


def unsome_(x: Some[T]) -> Optional[list[T]]:
    if x is None:
        return None
    if isinstance(x, list):
        return x
    return [x]


def unsome(x: Some[T]) -> list[T]:
    return unsome_(x) or []


def allow_singular(field: str) -> Any:
    return validator(field, pre=True, allow_reuse=True)(lambda cls, x: unsome_(x))


def allow_singular_none(field: str) -> Any:
    return validator(field, pre=True, allow_reuse=True)(lambda cls, x: unsome(x))


def intercalate(delimiter: T, items: Iterable[Iterable[T]]) -> list[T]:
    it = iter(items)
    try:
        result = list(next(it))
    except StopIteration:
        return []
    for x in it:
        result.append(delimiter)
        result.extend(list(x))
    return result


RunArg = Union[str, Path]


async def run(*args: RunArg) -> subprocess.CompletedProcess:
    return await trio.run_process(args, check=True)


async def run_ok(*args: RunArg) -> bool:
    try:
        result = await trio.run_process(
            args, check=False, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        return result.returncode == 0
    except FileNotFoundError:
        return False


async def run_output(*args: RunArg, silent: bool = False) -> str:
    result = await trio.run_process(
        args, capture_stdout=True, capture_stderr=silent, check=True
    )
    return result.stdout.decode().strip()


def flatten(items: Iterable[Iterable[T]]) -> list[T]:
    return [item for sublist in items for item in sublist]


def url_version(url: str) -> str:
    head_response = requests.head(url, allow_redirects=True)
    return head_response.headers["etag"]


def raise_(exception: BaseException) -> Any:
    raise exception


AsyncRet = Coroutine[Any, Any, T]


@overload
def async_cached(fn: Callable[[], AsyncRet[T]]) -> Callable[[], AsyncRet[T]]:
    ...


@overload
def async_cached(fn: Callable[[U], AsyncRet[T]]) -> Callable[[U], AsyncRet[T]]:
    ...


@overload
def async_cached(fn: Callable[[U, V], AsyncRet[T]]) -> Callable[[U, V], AsyncRet[T]]:
    ...


def async_cached(fn: Callable[..., AsyncRet[T]]) -> Callable[..., AsyncRet[T]]:
    return _async_cached_lock(None, fn)


@overload
def async_cached_lock(fn: Callable[[], AsyncRet[T]]) -> Callable[[], AsyncRet[T]]:
    ...


@overload
def async_cached_lock(fn: Callable[[U], AsyncRet[T]]) -> Callable[[U], AsyncRet[T]]:
    ...


@overload
def async_cached_lock(
    fn: Callable[[U, V], AsyncRet[T]]
) -> Callable[[U, V], AsyncRet[T]]:
    ...


def async_cached_lock(fn: Callable[..., AsyncRet[T]]) -> Callable[..., AsyncRet[T]]:
    return _async_cached_lock(trio.Lock(), fn)


class NoLock:
    async def __aenter__(self) -> None:
        pass

    async def __aexit__(self, *args: Any) -> None:
        pass


NO_LOCK = NoLock()


def _async_cached_lock(
    lock: Optional[trio.Lock], fn: Callable[..., AsyncRet[T]]
) -> Callable[..., AsyncRet[T]]:
    cache: dict[Any, T] = {}

    @wraps(fn)
    async def cached_fn(*args: Any) -> T:
        if args not in cache:
            async with lock or NO_LOCK:
                # Check again, someone holding the lock might have updated the cache
                if args not in cache:
                    cache[args] = await fn(*args)
        return cache[args]

    return cached_fn



import shlex

import requests
from pydantic import Field

from ..utils import allow_singular_none
from .manual_version import ManualVersion
from .root import Root
from .tracked import Tracked, Tracker


class NpmPackage(Root, ManualVersion, Tracked):
    package: str = Field(..., alias="npm")

    binaries: list[str] = Field(default_factory=list, alias="binary")
    binaries_val = allow_singular_none("binaries")

    @property
    def name(self) -> str:
        return self.package

    async def get_remote_version(self) -> str:
        result = requests.get(f"https://registry.npmjs.com/{self.package}")
        result.raise_for_status()

        details = result.json()
        return details["dist-tags"]["latest"]

    async def install_tracked(self, *, tracker: Tracker) -> None:
        args = ["npm", "exec", "--yes", "--package", self.package, "--"]

        await self.driver.run(*args, "true")

        npx_paths = await self.driver.run_output(*args, "echo $PATH")
        npx_path = next((path for path in npx_paths.split(":") if "_npx" in path), None)
        if not npx_path:
            raise Exception(f"Could not find npx path in {npx_paths}.")

        for name in self.binaries:
            target = await self.local() / "bin" / name
            await self.driver.write_file(
                target,
                f'#!/bin/sh\nPATH={shlex.quote(npx_path)}:$PATH\nexec "{name}" "$@"',
            )
            await self.driver.make_executable(target)
            tracker.track(target)

        await super().install_tracked(tracker=tracker)



from typing import Optional

from pydantic import Field

from ..utils import RunArg, async_cached, run_output
from .destination import Destination
from .tracked import Tracked, Tracker


class Clone(Destination, Tracked):
    repo: str = Field(..., alias="clone")

    @property
    def name(self) -> str:
        return self.repo

    async def directory_exists(self) -> bool:
        return await self.driver.run_ok("test", "-d", await self.destination())

    @async_cached
    async def git_args(self) -> list[RunArg]:
        return ["git", "-C", await self.destination()]

    async def run_git(self, *args: RunArg) -> None:
        await self.driver.run(*await self.git_args(), *args)

    async def local_version(self) -> Optional[str]:
        if not await self.directory_exists():
            return None
        return await self.driver.run_output(*await self.git_args(), "rev-parse", "HEAD")

    @property
    def remote(self):
        if self.repo.startswith("https://") or "@" in self.repo:
            return self.repo
        return f"https://github.com/{self.repo}.git"

    async def get_remote_version(self) -> str:
        return (await run_output("git", "ls-remote", self.remote, "HEAD")).split()[0]

    async def install_tracked(self, *, tracker: Tracker) -> None:
        destination = await self.destination()

        await self.driver.makedirs(destination.parent)

        if not await self.directory_exists():
            await self.driver.run("git", "clone", self.remote, destination)

        await self.run_git("remote", "set-url", "origin", self.remote)
        await self.run_git("fetch")
        default_branch = (
            await self.driver.run_output(
                *await self.git_args(), "rev-parse", "--abbrev-ref", "origin/HEAD"
            )
        ).split("/")[1]
        await self.run_git("switch", default_branch)
        await self.run_git("reset", "--hard", f"origin/{default_branch}")

        tracker.track(destination)

        await super().install_tracked(tracker=tracker)



import re
from typing import Callable, Iterator

from pydantic import BaseModel, Field

from .utils import T, allow_singular_none


class Filters(BaseModel):
    @staticmethod
    def includes_(substring: str) -> Callable[[str], bool]:
        return lambda x: substring in x.lower()

    @staticmethod
    def excludes_(substring: str) -> Callable[[str], bool]:
        return lambda x: substring not in x

    @staticmethod
    def startswith(prefix: str) -> Callable[[str], bool]:
        return lambda x: x.startswith(prefix)

    @staticmethod
    def endswith(suffix: str) -> Callable[[str], bool]:
        return lambda x: x.endswith(suffix)

    @staticmethod
    def regex_(regex: str) -> Callable[[str], bool]:
        regex_compiled = re.compile(regex)

        return lambda x: regex_compiled.match(x) is not None

    prefixes: list[str] = Field(default_factory=list, alias="prefix")
    prefixes_val = allow_singular_none("prefixes")

    suffixes: list[str] = Field(default_factory=list, alias="suffix")
    suffixes_val = allow_singular_none("suffixes")

    includes: list[str] = Field(default_factory=list, alias="include")
    includes_val = allow_singular_none("includes")

    excludes: list[str] = Field(default_factory=list, alias="exclude")
    excludes_val = allow_singular_none("excludes")

    regex: list[str] = Field(default_factory=list)
    regex_val = allow_singular_none("regex")

    def filters(self) -> Iterator[Callable[[str], bool]]:
        for prefix in self.prefixes:
            yield self.startswith(prefix)
        for suffix in self.suffixes:
            yield self.endswith(suffix)
        for include in self.includes:
            yield self.includes_(include)
        for exclude in self.excludes:
            yield self.excludes_(exclude)
        for regex in self.regex:
            yield self.regex_(regex)


def choose(candidates: list[T], filters: Iterator[Callable[[T], bool]]) -> T:
    if len(candidates) == 0:
        raise ValueError("No candidates to choose from.")
    while len(candidates) > 1:
        try:
            filter_fn = next(filters)
        except StopIteration:
            break

        new_candidates = list(filter(filter_fn, candidates))
        if new_candidates:
            candidates = new_candidates

    if len(candidates) == 1:
        return candidates[0]

    raise ValueError(f"Cannot choose between: {candidates}.")



from dataclasses import dataclass
from pathlib import Path

from .base import storage


@dataclass(frozen=True)
class InstalledFile:
    package: str
    path: str
    root: bool

    @property
    def path_(self) -> Path:
        return Path(self.path)


INSTALLED_FILES = storage("installed_file", InstalledFile)



import json

from pydantic import Field, validator

from .pip_base import PIP, PipBasePackage


class PipPackage(PipBasePackage):
    package: str = Field(..., alias="pip")

    @validator("package")
    def package_to_lower(cls, value: str) -> str:  # pylint: disable=no-self-argument
        return value.lower()

    async def get_all_versions(self) -> dict[str, str]:
        packages_json = await self.driver.run_output(
            *PIP, "list", "--format", "json", silent=True
        )

        # https://github.com/pypa/pip/issues/11282
        packages_json = packages_json.split("\n\n[notice]", maxsplit=1)[0]

        packages = json.loads(packages_json)

        return {package["name"].lower(): package["version"] for package in packages}

    async def install(self) -> None:
        await self.driver.run(*PIP, "install", "--upgrade", "--user", self.package)
        await super().install()



from abc import ABC, abstractmethod
from functools import cached_property
from typing import Optional

from pydantic import BaseModel, Field

from ..driver import Driver
from ..parallel import gather_
from ..state import DB
from ..utils import allow_singular, async_cached


class Package(BaseModel, ABC):
    class Config:
        allow_mutation = False
        arbitrary_types_allowed = True
        keep_untouched = (cached_property,)

    def __hash__(self) -> int:
        """Allow hashing to pass when gathering async results."""
        return hash(id(self))

    os: Optional[list[str]] = None
    os_val = allow_singular("os")

    distribution: Optional[list[str]] = None
    distribution_val = allow_singular("distribution")

    db: DB
    driver_: Driver = Field(..., alias="driver")

    @property
    def driver(self) -> Driver:
        return self.driver_

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    async def get_remote_version(self) -> str:
        pass

    @async_cached
    async def remote_version(self) -> str:
        return await self.get_remote_version()

    @abstractmethod
    async def local_version(self) -> Optional[str]:
        pass

    async def is_installed(self) -> bool:
        remote, local = await gather_(self.remote_version, self.local_version)
        return remote == local

    @abstractmethod
    async def install(self) -> None:
        pass

    async def applicable(self) -> bool:
        def check_os(name: str) -> bool:
            return self.os is None or name in self.os

        os = await self.driver.os()
        return os.switch_(
            linux=lambda linux: check_os("linux")
            and (self.distribution is None or linux.distribution in self.distribution),
            macos=check_os("darwin"),
        )

    async def ensure(self) -> bool:
        if not await self.applicable():
            return False
        if await self.is_installed():
            return False
        await self.install()
        return True



from contextlib import asynccontextmanager
from dataclasses import dataclass
from functools import partial
from typing import AsyncIterator, Awaitable, Callable, Generic, Optional, Union

import tqdm
import trio

from .utils import T, U, raise_


@dataclass
class PartialSuccess(Generic[T]):
    result: T


@dataclass
class PartialException:
    exception: BaseException


PartialResult = Union[PartialException, PartialSuccess[T]]


class PartialResults(Exception, Generic[T]):
    def __init__(self, results: list[PartialResult[T]]):
        self.results = results


TERMINAL_LOCK = trio.Lock()

CURRENT_TQDM: Optional[tqdm.tqdm] = None


async def parallel_map_tqdm(
    action: Callable[[T], Awaitable[U]], items: list[T]
) -> list[U]:
    with tqdm.tqdm(total=len(items)) as progress:
        global CURRENT_TQDM  # pylint:disable=global-statement
        async with TERMINAL_LOCK:
            CURRENT_TQDM = progress
        try:

            async def action_and_update(item: T) -> U:
                result = await action(item)
                async with TERMINAL_LOCK:
                    progress.update(1)
                return result

            return await gather_(*(partial(action_and_update, item) for item in items))
        finally:
            async with TERMINAL_LOCK:
                CURRENT_TQDM = None


@asynccontextmanager
async def parallel_map_pause() -> AsyncIterator[None]:
    async with TERMINAL_LOCK:
        progress = CURRENT_TQDM
        if progress:
            progress.clear()
        try:
            yield
        finally:
            if progress:
                progress.refresh()


async def gather(*tasks: Callable[[], Awaitable[T]]) -> list[PartialResult[T]]:
    async def collect(
        index: int,
        task: Callable[[], Awaitable[T]],
        results: dict[int, PartialResult[T]],
    ):
        try:
            results[index] = PartialSuccess(result=await task())
        except BaseException as e:  # pylint:disable=broad-exception-caught
            results[index] = PartialException(exception=e)

    results: dict[int, PartialResult[T]] = {}

    async with trio.open_nursery() as nursery:
        for index, task in enumerate(tasks):
            nursery.start_soon(collect, index, task, results)
    return [results[i] for i in range(len(tasks))]


async def gather_(*tasks: Callable[[], Awaitable[T]]) -> list[T]:
    partial_results = await gather(*tasks)

    return [
        result.result
        if isinstance(result, PartialSuccess)
        else raise_(PartialResults(partial_results))
        for result in partial_results
    ]



import re
from abc import ABC, abstractmethod
from typing import Optional, cast

import requests

from .base import Package

PIP = ["python3", "-m", "pip"]


class PipBasePackage(Package, ABC):
    package: str

    @property
    def name(self) -> str:
        return self.package

    @abstractmethod
    async def get_all_versions(self) -> dict[str, str]:
        pass

    async def local_version(self) -> Optional[str]:
        return (await self.get_all_versions()).get(self.package)

    async def _get_pypi_version(self) -> Optional[str]:
        pypi_info = requests.get(f"https://pypi.org/pypi/{self.package}/json").json()
        try:
            return pypi_info["info"]["version"]
        except KeyError:
            return None

    async def _get_index_version(self) -> Optional[str]:
        check = await self.driver.run_(
            *PIP,
            "index",
            "versions",
            self.package,
            check=False,
            silent=True,
            capture_output=True,
        )
        if not check.ok:
            return None
        output = cast(str, check.output)
        version = re.search(r"\(([^)]+)\)", output)
        if not version:
            raise Exception(f"Cannot parse pip output: {output}")
        return version[1]

    async def get_remote_version(self) -> str:
        if version := await self._get_pypi_version():
            return version

        if version := await self._get_index_version():
            return version

        raise Exception(f"Cannot find latest version of package '{self.package}'.")



from abc import ABC
from functools import cached_property
from pathlib import Path

from ..driver import Driver
from ..utils import async_cached
from .base import Package


class Root(Package, ABC):
    root: bool = False

    @cached_property
    def driver(self) -> Driver:
        return super().driver.with_root(self.root)

    @async_cached
    async def local(self) -> Path:
        if self.root:
            return Path("/usr/local")
        else:
            return await self.driver.local()



import dataclasses
import sqlite3
import threading
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterable,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from uuid import uuid4

if TYPE_CHECKING:
    from _typeshed import DataclassInstance  # pragma: no cover

    M = TypeVar("M", bound=DataclassInstance)  # pragma: no cover
else:
    M = TypeVar("M")

DB_PATH = "state.sqlite"


class DB:
    """Thread-safe SQLite database connection."""

    def __init__(self, path: Union[str, Path]) -> None:
        self.path = path
        self.cache = threading.local()

    @property
    def instance(self) -> sqlite3.Connection:
        try:
            return self.cache.connection
        except AttributeError:
            pass

        connection = sqlite3.connect(self.path)
        connection.row_factory = sqlite3.Row
        self.cache.connection = connection
        return connection


class Storage(Generic[M], metaclass=ABCMeta):
    @abstractmethod
    def __getitem__(self, key: str) -> M:
        pass

    @abstractmethod
    def __delitem__(self, key: str) -> None:
        pass

    @abstractmethod
    def __setitem__(self, key: str, value: M) -> None:
        pass

    @abstractmethod
    def append(self, value: M) -> None:
        pass

    @abstractmethod
    def find_ids(self, **kwargs: Any) -> Iterable[tuple[str, M]]:
        pass

    def find(self, **kwargs: Any) -> Iterable[M]:
        for _, value in self.find_ids(**kwargs):
            yield value

    @abstractmethod
    def delete(self, **kwargs: Any) -> None:
        pass


StorageDefinition = Callable[[DB], Storage[M]]


def storage(name: str, klass: Type[M]) -> StorageDefinition[M]:
    attributes: list[str] = [field.name for field in dataclasses.fields(klass)]

    def storage_table(db: DB) -> Storage[M]:
        db.instance.execute(
            f'CREATE TABLE IF NOT EXISTS {name} (id TEXT PRIMARY KEY, {", ".join(attributes)})'
        )

        class StorageImpl(Storage[M]):
            def where_clause(self, **kwargs: Any) -> Tuple[str, list[Any]]:
                result = "1=1"
                values = []
                for key, value in kwargs.items():
                    result += f" AND {key} = ?"
                    values.append(value)
                return result, values

            def find_ids(self, **kwargs: Any) -> Iterable[tuple[str, M]]:
                where, values = self.where_clause(**kwargs)

                cursor = db.instance.execute(
                    f"SELECT * FROM {name} WHERE {where}", values
                )

                while rows := cursor.fetchmany():
                    for row in rows:
                        attributes = {
                            key: row[key] for key in row.keys() if key != "id"
                        }
                        yield row["id"], klass(**attributes)

            def __getitem__(self, key: str) -> M:
                items = self.find(id=key)
                try:
                    return next(iter(items))
                except StopIteration:
                    raise KeyError(key) from None

            def __delitem__(self, key: str) -> None:
                self.delete(id=key)

            def delete(self, **kwargs: Any) -> None:
                where, values = self.where_clause(**kwargs)
                db.instance.execute(f"DELETE FROM {name} WHERE {where}", values)

            def __setitem__(self, key: str, value: M) -> None:
                with db.instance:
                    del self[key]
                    attr_clause = "?"
                    attr_values = [key]
                    for attribute in attributes:
                        attr_clause += ", ?"
                        attr_values.append(getattr(value, attribute))
                    db.instance.execute(
                        f"INSERT INTO {name} VALUES ({attr_clause})", attr_values
                    )

            def append(self, value: M) -> None:
                self[str(uuid4())] = value

        return StorageImpl()

    return storage_table



import json

from pydantic import Field, validator

from .pip_base import PipBasePackage

PIPX = ["python3", "-m", "pipx"]


class PipxPackage(PipBasePackage):
    package: str = Field(..., alias="pipx")

    @validator("package")
    def package_to_lower(cls, value: str) -> str:  # pylint: disable=no-self-argument
        return value.lower()

    async def get_all_versions(self) -> dict[str, str]:
        pipx_list = json.loads(
            await self.driver.run_output(*PIPX, "list", "--json", silent=True)
        )
        packages = (
            item["metadata"]["main_package"] for item in pipx_list["venvs"].values()
        )
        return {package["package"]: package["package_version"] for package in packages}

    async def install(self) -> None:
        cmd = "install" if await self.local_version() is None else "upgrade"
        await self.driver.run(*PIPX, cmd, self.package)
        await super().install()
